# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

server_account = 'sixtadm'

from .tools import timestamp, log_debug, log_info, log_error, untar
from .fs import *
from .paths import *